# Copyright (c) 2019 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

NETWORK_PKEY = "pkey"

OPA_BINARY = "opafmvf"

# Operations
OPA_CREATE = "create"
OPA_ADD = "add"
OPA_REMOVE = "remove"
OPA_DELETE = "delete"
OPA_COMMIT = "commit"
OPA_ABORT = "abort"
OPA_EXIST = "exist"
OPA_RESTART = "restart"
OPA_ISMEMBER = "ismember"
OPA_ISNOTMEMBER = "isnotmember"
OPA_RESET = "reset"
OPA_RELOAD = "reload"

# Statuses
OPA_NF = 2
OPA_SUCCESS = 0
OPA_FAIL = 1

# States
OPA_COMPLETED = "completed"
OPA_FAILED = "failed"

ALLOWED_OPA_COMMANDS = (OPA_CREATE, OPA_ADD,
                        OPA_REMOVE, OPA_DELETE,
                        OPA_COMMIT, OPA_RESTART,
                        OPA_ISMEMBER, OPA_ISNOTMEMBER,
                        OPA_RESET, OPA_RELOAD)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import model_base
from neutron_lib.db import sqlalchemytypes
from oslo_utils import timeutils
import sqlalchemy as sa


class OmniPathRevisionNumbers(model_base.BASEV2):
    __tablename__ = 'omnipath_revision_journal'
    __table_args__ = (
        model_base.BASEV2.__table_args__
    )
    standard_attr_id = sa.Column(
        sa.BigInteger().with_variant(sa.Integer(), 'sqlite'),
        sa.ForeignKey('standardattributes.id', ondelete='SET NULL'),
        unique=True)
    resource_uuid = sa.Column(sa.String(36), nullable=False, primary_key=True)
    resource_type = sa.Column(sa.String(36), nullable=False, primary_key=True)
    revision_number = sa.Column(
        sa.BigInteger().with_variant(sa.Integer(), 'sqlite'),
        default=0, nullable=False)
    data = sa.Column(sa.PickleType, nullable=True)
    created_at = sa.Column(sqlalchemytypes.TruncatedDateTime,
                           default=timeutils.utcnow, nullable=False)
    updated_at = sa.Column(sqlalchemytypes.TruncatedDateTime,
                           onupdate=timeutils.utcnow)
    state = sa.Column(sa.Enum('pending', 'failed', 'completed',
                              'waiting'),
                      nullable=False, default='pending')



# Copyright 2019 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


from neutron_lib.plugins import directory
from oslo_config import cfg

from neutron.plugins.ml2.drivers import type_flat  # noqa
from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron.tests.unit.testlib_api import SqlTestCaseLight
from neutron_lib import context

from omnipath.db import models


MECH_OMNIPATH = 'omnipath_mech'


class DBTestCase(SqlTestCaseLight):

    def setUp(self):
        super(DBTestCase, self).setUp()
        self.db_context = context.get_admin_context()
        self.session = self.db_context.session

    def tearDown(self):
        super(DBTestCase, self).tearDown()
        self.session.query(models.OmniPathRevisionNumbers).delete()


class TestOmniPathBase(test_plugin.Ml2PluginV2TestCase):

    _mechanism_drivers = [MECH_OMNIPATH]
    _extension_drivers = ['port_security']

    def setUp(self):
        cfg.CONF.set_override('extension_drivers',
                              self._extension_drivers,
                              group='ml2')
        cfg.CONF.set_override('tenant_network_types',
                              ['vlan'],
                              group='ml2')
        cfg.CONF.set_override('flat_networks',
                              ['public', 'mynet'],
                              group='ml2_type_flat')
        cfg.CONF.set_override('network_vlan_ranges',
                              ['mynet:2:2000'],
                              group='ml2_type_vlan')
        super(TestOmniPathBase, self).setUp()
        mechmanager = directory.get_plugin().mechanism_manager
        self.mech_driver = mechmanager.mech_drivers[MECH_OMNIPATH].obj



# Copyright (c) 2019 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from neutron_lib.api.definitions import portbindings
from neutron_lib.callbacks import resources
from neutron_lib import context as nl_context
from neutron_lib.plugins.ml2 import api
from oslo_log import log as logging

from neutron.db import provisioning_blocks
from neutron.objects import ports

from omnipath.common import constants as op_const
from omnipath.db import api as opadbapi
from omnipath.mechanism_driver import fabric_agent
from omnipath import omnipath_thread as ojournal
from omnipath import omnipath_worker

LOG = logging.getLogger(__name__)

# TODO(manjeets) use this parameter from config
sync_time = 10


class OmnipathMechanismDriver(api.MechanismDriver):
    def initialize(self):
        LOG.debug("OMNIPATH initializing omnipath...")
        self.opafmvf = fabric_agent.FabricAgentClient()
        # Runs sync() every "sync_time" seconds
        self.omnipath_thread = ojournal.OmniPathThread(
            fabric_cli=self.opafmvf.cli)
        self.omnipath_thread.start()
        self.supported_vnic_types = [portbindings.VNIC_NORMAL,
                                     portbindings.VNIC_DIRECT,
                                     "baremetal"]

    def get_workers(self):
        workers = [omnipath_worker.OmniPathPeriodicProcessor()]
        return workers

    @staticmethod
    def update_port_status_db(context, port_id, status):
        """Update Ports in DB

        :param port_id: ID of port to update status
        :param status: Status of the port
        :return: Updated Port Object
        """
        LOG.debug("OMNIPATH update_port_status_db...")
        ctx = context.get_admin_context()
        return ports.Port.update_object(
            ctx, {'status': status}, port_id=port_id)

    @staticmethod
    def _is_port_supported(port):
        """Return whether a port is supported by this driver.

        Ports supported by this driver have a VNIC type of 'baremetal'.
        :param port: The port to check
        :returns: Whether the port is supported by the NGS driver
        """
        vnic_type = port[portbindings.VNIC_TYPE]
        LOG.debug("OMNIPATH _is_port_supported(%s), result %s..." % (
            vnic_type, (vnic_type == portbindings.VNIC_BAREMETAL)))
        return vnic_type == portbindings.VNIC_BAREMETAL

    def create_network_precommit(self, context):
        """Allocate resources for a new network.

        :param context: NetworkContext instance describing the new
        network.

        Create a new network, allocating resources as necessary in the
        database. Called inside transaction context on session. Call
        cannot block.  Raising an exception will result in a rollback
        of the current transaction.
        """
        # (manjeets) add the revision number logic here.
        LOG.debug("OMNIPATH create_network_precommit...")
        net = context.current
        res_id = net['id']
        res_type = "network"
        net['operation'] = op_const.OPA_CREATE
        net['op_type'] = "config"
        opadbapi.record_pending_entry(
            context._plugin_context, res_id, res_type, net)

    def create_network_postcommit(self, context):
        """Create a network.

        :param context: NetworkContext instance describing the new
        network.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.
        """
        self.omnipath_thread.set_sync_event()

    def update_network_precommit(self, context):
        """Update resources of a network.

        :param context: NetworkContext instance describing the new
        state of the network, as well as the original state prior
        to the update_network call.

        Update values of a network, updating the associated resources
        in the database. Called inside transaction context on session.
        Raising an exception will result in rollback of the
        transaction.

        update_network_precommit is called for all changes to the
        network state. It is up to the mechanism driver to ignore
        state or state changes that it does not know or care about.
        """
        return

    def update_network_postcommit(self, context):
        """Update a network.

        :param context: NetworkContext instance describing the new
        state of the network, as well as the original state prior
        to the update_network call.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.

        update_network_postcommit is called for all changes to the
        network state.  It is up to the mechanism driver to ignore
        state or state changes that it does not know or care about.
        """
        # Is Network update required for OPA?
        return

    def delete_network_precommit(self, context):
        """Delete resources for a network.

        :param context: NetworkContext instance describing the current
        state of the network, prior to the call to delete it.

        Delete network resources previously allocated by this
        mechanism driver for a network. Called inside transaction
        context on session. Runtime errors are not expected, but
        raising an exception will result in rollback of the
        transaction.
        """
        net = context.current
        res_id = net['id']
        res_type = "network"
        net['operation'] = op_const.OPA_DELETE
        net['op_type'] = "config"
        row = opadbapi.get_resource_row(context._plugin_context,
                                        res_id, res_type)
        if not row:
            return
        row.data = net
        row.state = 'pending'
        context._plugin_context.session.merge(row)

    def delete_network_postcommit(self, context):
        """Delete a network.

        :param context: NetworkContext instance describing the current
        state of the network, prior to the call to delete it.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Runtime errors are not
        expected, and will not prevent the resource from being
        deleted.
        """
        self.omnipath_thread.set_sync_event()

    def create_subnet_precommit(self, context):
        """Allocate resources for a new subnet.

        :param context: SubnetContext instance describing the new
        subnet.
        rt = context.current
        device_id = port['device_id']
        device_owner = port['device_owner']
        Create a new subnet, allocating resources as necessary in the
        database. Called inside transaction context on session. Call
        cannot block.  Raising an exception will result in a rollback
        of the current transaction.
        """
        return

    def create_subnet_postcommit(self, context):
        """Create a subnet.

        :param context: SubnetContext instance describing the new
        subnet.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.
        """
        return

    def update_subnet_precommit(self, context):
        """Update resources of a subnet.

        :param context: SubnetContext instance describing the new
        state of the subnet, as well as the original state prior
        to the update_subnet call.

        Update values of a subnet, updating the associated resources
        in the database. Called inside transaction context on session.
        Raising an exception will result in rollback of the
        transaction.

        update_subnet_precommit is called for all changes to the
        subnet state. It is up to the mechanism driver to ignore
        state or state changes that it does not know or care about.
        """
        return

    def update_subnet_postcommit(self, context):
        """Update a subnet.

        :param context: SubnetContext instance describing the new
        state of the subnet, as well as the original state prior
        to the update_subnet call.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.

        update_subnet_postcommit is called for all changes to the
        subnet state.  It is up to the mechanism driver to ignore
        state or state changes that it does not know or care about.
        """
        return

    def delete_subnet_precommit(self, context):
        """Delete resources for a subnet.

        :param context: SubnetContext instance describing the current
        state of the subnet, prior to the call to delete it.

        Delete subnet resources previously allocated by this
        mechanism driver for a subnet. Called inside transaction
        context on session. Runtime errors are not expected, but
        raising an exception will result in rollback of the
        transaction.
        """
        return

    def delete_subnet_postcommit(self, context):
        """Delete a subnet.

        :param context: SubnetContext instance describing the current
        state of the subnet, prior to the call to delete it.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Runtime errors are not
        expected, and will not prevent the resource from being
        deleted.
        """
        return

    def create_port_precommit(self, context):
        """Allocate resources for a new port.

        :param context: PortContext instance describing the port.

        Create a new port, allocating resources as necessary in the
        database. Called inside transaction context on session. Call
        cannot block.  Raising an exception will result in a rollback
        of the current transaction.
        """
        LOG.debug("OMNIPATH create_port_precommit...")
        port = context.current
        port['operation'] = op_const.OPA_CREATE
        port['op_type'] = "config"
        profile = context.current.get(portbindings.PROFILE)
        node_guid = None
        LOG.debug("OMNIPATH port id %s has profile %s" % (port['id'], profile))
        if profile:
            node_guid = profile.get('guid')
        if not profile or not node_guid:
            # port doesn't belong to this backend or OPA info not yet available
            LOG.debug("OMNIPATH the port does not belong to this backend")
            return
        LOG.debug("OMNIPATH port id %s has node_guid %s" % (
            port['id'], node_guid))
        port['guid'] = node_guid
        opadbapi.record_pending_entry(context._plugin_context,
                                      port['id'], "port", port,
                                      state='waiting')
        LOG.debug("OMNIPATH added record for port %s 'waiting'" % port['id'])

    def create_port_postcommit(self, context):
        """Create a port.

        :param context: PortContext instance describing the port.

        Called after the transaction completes. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance.  Raising an exception will
        result in the deletion of the resource.
        """
        return

    def update_port_precommit(self, context):
        """Update resources of a port.

        :param context: PortContext instance describing the new
        state of the port, as well as the original state prior
        to the update_port call.

        Called inside transaction context on session to complete a
        port update as defined by this mechanism driver. Raising an
        exception will result in rollback of the transaction.

        update_port_precommit is called for all changes to the port
        state. It is up to the mechanism driver to ignore state or
        state changes that it does not know or care about.
        """
        port = context.current
        if 'dhcp' in port['device_owner']:
            return

    def update_port_postcommit(self, context):
        """Update a port.

        :param context: PortContext instance describing the new
        state of the port, as well as the original state prior
        to the update_port call.

        Called after the transaction completes. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance.  Raising an exception will
        result in the deletion of the resource.

        update_port_postcommit is called for all changes to the port
        state. It is up to the mechanism driver to ignore state or
        state changes that it does not know or care about.
        """
        # TODO(manjeets) update if needed
        return

    def _unplug_port_from_network(self, context):
        self.delete_port_precommit(context)
        self.omnipath_thread.set_sync_event()

    def delete_port_precommit(self, context):
        """Delete resources of a port.

        :param context: PortContext instance describing the current
        state of the port, prior to the call to delete it.

        Called inside transaction context on session. Runtime errors
        are not expected, but raising an exception will result in
        rollback of the transaction.
        """
        LOG.debug("OMNIPATH delete_port_precommit ...")
        port = context.current
        port_row = opadbapi.get_resource_row(
            context._plugin_context, port['id'], "port")
        portdict = port_row.get("data") if port_row else None
        node_guid = portdict.get('guid') if portdict else None
        if not port_row or not node_guid:
            return
        port["operation"] = op_const.OPA_DELETE
        port["guid"] = node_guid
        port_row.state = 'waiting'
        port_row.data = port
        context._plugin_context.session.merge(port_row)

    def delete_port_postcommit(self, context):
        """Delete a port.

        :param context: PortContext instance describing the current
        state of the port, prior to the call to delete it.

        Called after the transaction completes. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance.  Runtime errors are not
        expected, and will not prevent the resource from being
        deleted.
        """
        # Remove port from DB. Once removed, it will be
        # synced every "sync_time" as done in  __init__
        self.omnipath_thread.set_sync_event()

    def bind_port(self, context):
        """Attempt to bind a port.

        :param context: PortContext instance describing the port

        This method is called outside any transaction to attempt to
        establish a port binding using this mechanism driver. Bindings
        may be created at each of multiple levels of a hierarchical
        network, and are established from the top level downward. At
        each level, the mechanism driver determines whether it can
        bind to any of the network segments in the
        context.segments_to_bind property, based on the value of the
        context.host property, any relevant port or network
        attributes, and its own knowledge of the network topology. At
        the top level, context.segments_to_bind contains the static
        segments of the port's network. At each lower level of
        binding, it contains static or dynamic segments supplied by
        the driver that bound at the level above. If the driver is
        able to complete the binding of the port to any segment in
        context.segments_to_bind, it must call context.set_binding
        with the binding details. If it can partially bind the port,
        it must call context.continue_binding with the network
        segments to be used to bind at the next lower level.

        If the binding results are committed after bind_port returns,
        they will be seen by all mechanism drivers as
        update_port_precommit and update_port_postcommit calls. But if
        some other thread or process concurrently binds or updates the
        port, these binding results will not be committed, and
        update_port_precommit and update_port_postcommit will not be
        called on the mechanism drivers with these results. Because
        binding results can be discarded rather than committed,
        drivers should avoid making persistent state changes in
        bind_port, or else must ensure that such state changes are
        eventually cleaned up.

        Implementing this method explicitly declares the mechanism
        driver as having the intention to bind ports. This is inspected
        by the QoS service to identify the available QoS rules you
        can use with ports.
        """
        port_data = context.current
        LOG.debug("OMNIPATH bind_port(%s) ..." % port_data['id'])
        port = opadbapi.get_resource_row(context._plugin_context,
                                         port_data['id'], "port")

        profile = port_data.get('binding:profile')
        node_guid = None
        if profile:
            node_guid = profile.get('guid')

        extra_dhcp_opts = port_data.get('extra_dhcp_opts')
        tmp_node = None
        for option in extra_dhcp_opts:
            if option.get("opt_name") == "61":
                tmp_node = option.get("opt_value")
                LOG.debug("OMNIPATH bind_port: found GUID=%s" % tmp_node)

        if not tmp_node:
            LOG.debug("OMNIPATH No guid provided over extra_dhcp_opt...")
        else:
            node_guid = tmp_node[-16:]
            LOG.debug("OMNIPATH Set guid from extra_dhcp_opt: %s", node_guid)

        if not port_data or not profile or not node_guid:
            LOG.debug("OMNIPATH No guid provided for this port, skipping the "
                      "binding")
            return

        if not port:
            LOG.debug("OMNIPATH late-inserting revision journal entry")
            port = port_data
            # FIXME(igordc): the below is only a workaround for now,
            #                the thread needs to do this instead
            opadbapi.record_pending_entry(context._plugin_context,
                                          port['id'], "port", port_data,
                                          state='waiting')

        port_data["operation"] = op_const.OPA_ADD
        port_data['guid'] = node_guid
        # context._plugin_context.session.merge(port)
        net_id = context.network.current['id']
        LOG.debug("OMNIPATH Attempting to bind port %(port)s on "
                  "network %(network)s",
                  {'port': context.current['id'],
                   'network': net_id})
        segment = context.segments_to_bind[0]
        LOG.debug("OMNIPATH segment %s" % segment)
        if not segment:
            LOG.error("OMNIPATH Port Binding error no valid segments to bind")
            return
        context.set_binding(segment[api.ID], "PKEY", {}, status='ACTIVE')
        provisioning_blocks.provisioning_complete(
            nl_context.get_admin_context(),
            port_data['id'], resources.PORT,
            provisioning_blocks.L2_AGENT_ENTITY)
        self.omnipath_thread.set_sync_event()



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from omnipath.db import models


def record_pending_entry(context, res_uuid,
                         res_type, data,
                         state='pending'):
    row = models.OmniPathRevisionNumbers(
        resource_uuid=res_uuid,
        resource_type=res_type,
        data=data,
        state=state)
    context.session.add(row)


def get_resource_row(context, res_uuid, res_type):
    res = context.session.query(models.OmniPathRevisionNumbers).filter_by(
        resource_uuid=res_uuid, resource_type=res_type).first()
    return res


def get_all_entries_by_state(context, state):
    all_entries = context.session.query(
        models.OmniPathRevisionNumbers).filter_by(
            state=state)
    return all_entries.all()


def update_row_state(context, state, row):
    row.state = state
    context.session.merge(row)


def update_multiple_rows(context, state, res_ids):
    all_entries = context.session.query(
        models.OmniPathRevisionNumbers).filter(
            models.OmniPathRevisionNumbers.resource_uuid.in_(
                res_ids))
    for row in all_entries.all():
        update_row_state(context, state, row)


def mark_entry_complete(context, row, status):
    if row:
        row.update(status='complete')



# Copyright (c) 2019 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import threading

import six

from neutron_lib import context as nl_context
from oslo_log import log as logging

from omnipath.common import constants as const
from omnipath.db import api as opadbapi
from omnipath.mechanism_driver import fabric_agent
from omnipath import omnipath_exceptions as omp_exc

LOG = logging.getLogger(__name__)

# TODO(manjeets) use this parameter from config
sync_time = 10


@six.add_metaclass(abc.ABCMeta)
class OmniPathThread(object):

    def __init__(self, start_thread=True, fabric_cli=None):
        self.event = threading.Event()
        self._sync_time = 300
        self.status = False
        self._omni_sync_thread = self._create_omni_sync_thread()
        self._omni_sync_thread_stop = threading.Event()
        self.fabric_cli = fabric_cli if fabric_cli else (
            fabric_agent.FabricAgentClient().cli)
        if start_thread:
            self.start()

    def _create_omni_sync_thread(self):
        return threading.Thread(name='sync', target=self.run_sync)

    def start(self):
        LOG.debug("OMNIPATH/THREAD Starting a new omnipath sync thread")
        if self._omni_sync_thread_stop.is_set():
            self._omni_sync_thread_stop.clear()
            self._omni_sync_thread = self._create_omni_sync_thread()

        if not self._omni_sync_thread.is_alive():
            self._omni_sync_thread.start()

    def stop(self, timeout=None):
        LOG.debug("OMNIPATH/THREAD Stopping the omnipath sync thread")
        if self._omni_sync_thread.is_alive():
            self._omni_sync_thread_stop.set()
            self.set_sync_event()
            self._omni_sync_thread.join(timeout)
            self.fabric_cli.osfa_management_commands("abort")

    def set_sync_event(self):
        self.event.set()

    def run_sync(self):
        while not self._omni_sync_thread_stop.is_set():
            try:
                LOG.debug("OMNIPATH/THREAD run_sync() cycle")
                self.event.wait()
                self.event.clear()
                self.sync_omnipath_operations()
            except Exception:
                LOG.exception("OMNIPATH/THREAD Error on running omnipath "
                              "sync run_sync")

    def _commit_and_reload(self):
        try:
            LOG.debug("OMNIPATH/THREAD opafmvf commit+reload")
            self.fabric_cli.osfa_management_commands(const.OPA_COMMIT)
            self.fabric_cli.osfa_management_commands(const.OPA_RELOAD)
        except Exception:
            LOG.exception("OMNIPATH/THREAD Error on omnipath sync check if "
                          "fabric is up")
            raise omp_exc.FabricAgentCLIError

    def sync_omnipath_operations(self):
        LOG.debug("OMNIPATH/THREAD Started syncing with omnipath fabric")
        try:
            context = nl_context.get_admin_context()
            all_pending_entries = opadbapi.get_all_entries_by_state(
                context, 'pending')
            all_waiting_ports = opadbapi.get_all_entries_by_state(
                context, 'waiting')
            if all_pending_entries:
                for entry in all_pending_entries:
                    try:
                        self._process_entry(context, entry)
                    except Exception:
                        LOG.error(
                            "OMNIPATH/THREAD Error on syncing omnipath entry")
            if all_waiting_ports:
                updated_ports = self._process_port_batch(
                    context, all_waiting_ports)
                if updated_ports > 0:
                    self._commit_and_reload()
        except Exception:
            LOG.error(
                "OMNIPATH/THREAD Error on omnipath sync check if fabric is up")

    def _prepare_ports_batch(self, all_waiting_ports):
        LOG.debug("OMNIPATH/THREAD _prepare_ports_batch()")
        port_add_batch = {}
        port_del_batch = {}
        for port in all_waiting_ports:
            network_id = port.data['network_id']
            if port.data['operation'] == const.OPA_ADD:
                if network_id not in port_add_batch.keys():
                    port_add_batch[network_id] = [(port.data['guid'],
                                                   port.resource_uuid)]
                else:
                    port_add_batch[network_id].append((port.data['guid'],
                                                       port.resource_uuid))
            elif port.data['operation'] == const.OPA_DELETE:
                if network_id not in port_del_batch.keys():
                    port_del_batch[network_id] = [(port.data['guid'],
                                                   port.resource_uuid)]
                else:
                    port_del_batch[network_id].append((port.data['guid'],
                                                      port.resource_uuid))
        return port_add_batch, port_del_batch

    def _process_port_batch(self, context, all_waiting_ports):
        port_add_batch, port_del_batch = self._prepare_ports_batch(
            all_waiting_ports)
        port_bind_ids = []
        port_del_ids = []
        for network in port_add_batch.keys():
            port_args = ["0x{}".format(x[0]) for x in port_add_batch[network]]
            port_guids = " ".join(port_args)
            self.fabric_cli.osfa_config_commands(
                const.OPA_ADD, network, port_guids)
            port_bind_ids = [x[1] for x in port_add_batch[network]]
            opadbapi.update_multiple_rows(context, "completed",
                                          port_bind_ids)

        LOG.debug("OMNIPATH/THREAD all_waiting_ports=%s" % all_waiting_ports)
        for network in port_del_batch.keys():
            port_args = ["0x{}".format(x[0]) for x in port_del_batch[network]]
            port_guids = " ".join(port_args)
            self.fabric_cli.osfa_config_commands(
                const.OPA_REMOVE, network, port_guids)
            port_del_ids = [x[1] for x in port_del_batch[network]]
            opadbapi.update_multiple_rows(context, "completed",
                                          port_del_ids)

        return len(port_bind_ids) + len(port_del_ids)

    def _process_entry(self, context, entry):
        LOG.debug("OMNIPATH/THREAD _process_entry()")
        data = entry.get('data')
        if entry.get('resource_type') == "network":
            net_id = data['id']
            if data['operation'] == const.OPA_CREATE:
                pkey = data['provider:segmentation_id']
                status = self.fabric_cli.osfa_config_commands(
                    const.OPA_CREATE, net_id, pkey)
                op_state = const.OPA_COMPLETED if status == 0 else (
                    const.OPA_FAILED)
                opadbapi.update_row_state(
                    context, const.OPA_COMPLETED, entry)
            elif data['operation'] == const.OPA_DELETE:
                status = self.fabric_cli.osfa_config_commands(
                    const.OPA_DELETE, net_id)
                op_state = const.OPA_COMPLETED if status == 0 else (
                    const.OPA_FAILED)
                self._commit_and_reload()
            opadbapi.update_row_state(context, op_state, entry)
        elif entry.get('resource_type') == 'port':
            net_id = data['network_id']
            port_guid = data['port_guid']
            LOG.debug(
                "OMNIPATH/THREAD adding port guid %s from pending" % port_guid)
            self.fabric_cli.osfa_config_commands(
                const.OPA_ADD, net_id, "0x{}".format(port_guid))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from neutron.db.migration.alembic_migrations import external
from neutron.db.migration import cli as migration
from neutron.db.migration.models import head
from neutron.tests.functional.db import test_migrations
from neutron.tests.unit import testlib_api

from omnipath.db import models  # noqa

# EXTERNAL_TABLES should contain all names of tables that are not related to
# current repo.

EXTERNAL_TABLES = external.TABLES

VERSION_TABLE = 'omnipath_alembic_version'


class _TestModelsMigrationsOmniPath(test_migrations._TestModelsMigrations):
    def db_sync(self, engine):
        cfg.CONF.set_override('connection', engine.url, group='database')
        for conf in migration.get_alembic_configs():
            self.alembic_config = conf
            self.alembic_config.neutron_config = cfg.CONF
            migration.do_alembic_command(conf, 'upgrade', 'heads')

    def get_metadata(self):
        return head.model_base.BASEV2.metadata

    def include_object(self, object_, name, type_, reflected, compare_to):
        if type_ == 'table' and (name.startswith('alembic') or
                                 name == VERSION_TABLE or
                                 name in EXTERNAL_TABLES):
            return False
        if type_ == 'index' and reflected and name.startswith("idx_autoinc_"):
            return False
        return True


class TestModelsMigrationsMysql(testlib_api.MySQLTestCaseMixin,
                                _TestModelsMigrationsOmniPath,
                                testlib_api.SqlTestCase):

    # Unstable test due to "bug 1687027"
    def test_models_sync(self):
        pass


class TestModelsMigrationsPostgresql(testlib_api.PostgreSQLTestCaseMixin,
                                     _TestModelsMigrationsOmniPath,
                                     testlib_api.SqlTestCase):
    pass



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import random

import mock
import paramiko

from oslo_config import cfg as config
from oslo_utils import uuidutils

from omnipath.mechanism_driver import fabric_agent
from omnipath.tests.unit import base


class TestOmniPathFabricAgent(base.TestOmniPathBase):

    def setUp(self):
        super(TestOmniPathFabricAgent, self).setUp()
        self.fabric_agent = self.mech_driver.opafmvf.cli

    @mock.patch.object(paramiko.SSHClient, 'set_missing_host_key_policy')
    @mock.patch.object(paramiko.SSHClient, 'connect')
    @mock.patch.object(paramiko.RSAKey, 'from_private_key_file')
    def test_connect(self, mock_key, mock_conn, mock_set):
        self._agent_hostname = config.CONF.ml2_omnipath.ip_address
        mock_key.return_value = 'key'
        self.fabric_agent.connect()
        mock_conn.assert_called_with(self._agent_hostname, port=22,
                                     username=None, pkey='key')

    def test_fabric_agent_management_command_reset(self):
        with mock.patch.object(self.fabric_agent,
                               "execute_command",
                               return_value=0) as mock_fa:
            self.fabric_agent.osfa_management_commands("reset")
            mock_fa.assert_called_once_with(" opafmvf reset")

    def test_fabric_agent_management_command_abort(self):
        with mock.patch.object(self.fabric_agent,
                               "execute_command",
                               return_value=0) as mock_fa:
            self.fabric_agent.osfa_management_commands("abort")
            mock_fa.assert_called_with(" killall -9 opafmvf")

    def test_fabric_agent_config_command_abort(self):
        with mock.patch.object(self.fabric_agent,
                               "execute_command",
                               return_value=0) as mock_fa:
            fake_arg = random.randint(0x0000, 0x7FFF)
            cmd = " opafmvf create fake_vf --pkey "
            self.fabric_agent.osfa_config_commands("create",
                                                   "fake_vf",
                                                   fake_arg)
            mock_fa.assert_called_with(cmd + str(fake_arg))

    def test_exceptions(self):
        with mock.patch.object(fabric_agent.LOG, 'error') as mock_log:
            fake_arg = random.randint(0x0000, 0x7FFF)
            self.fabric_agent.osfa_management_commands("hello")
            mock_log.assert_called()
            self.fabric_agent.osfa_query_commands("hello"
                                                  "fake_vf",
                                                  "fake_arg")
            mock_log.assert_called()
            self.fabric_agent.osfa_config_commands("asdf",
                                                   "fake_vf",
                                                   fake_arg)
            mock_log.assert_called()

    def test_fabric_agent_queries(self):
        with mock.patch.object(self.fabric_agent,
                               "execute_command",
                               return_value=0) as mock_fa:
            command = " opafmvf ismember fake_vf "
            self.fabric_agent.osfa_query_commands("ismember",
                                                  "fake_vf")
            mock_fa.assert_called_once_with(command)

    def test_get_port_status(self):
        with mock.patch.object(self.fabric_agent,
                               "execute_command",
                               return_value=0):
            fake_guid = uuidutils.generate_uuid()
            fake_vf = "fake_vf"
            self.assertEqual(
                self.mech_driver.opafmvf.get_port_status(
                    fake_vf, fake_guid), ("UP"))



# Copyright (c) 2019 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg
from oslo_log import log as logging

from neutron._i18n import _

LOG = logging.getLogger(__name__)

opt_group = cfg.OptGroup(name="ml2_omnipath",
                         title="Omnipath configuration")

omnipath_opts = [
    cfg.StrOpt(
        "ip_address",
        default="localhost",
        help=_("Host IP Address of the OPA FM Agent")),
    cfg.StrOpt(
        "username",
        help=_("Username of the OPA FM Agent")),
    cfg.StrOpt(
        "ssh_key",
        help=_("Private key path to access OPA FM Agent")),
    cfg.StrOpt(
        "poll_interval",
        help=_("Interval in seconds which a full sync is done"
               "with OPA FM Agent ")),
]


cfg.CONF.register_opts(omnipath_opts, "ml2_omnipath")


def list_opts():
    return [('ml2_omnipath', omnipath_opts)]



#    Copyright (c) 2019 Intel Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import mock

from omnipath.db import api as opadbapi
from omnipath.mechanism_driver import fabric_agent
from omnipath import omnipath_thread
from omnipath.tests.unit.mechanism_driver import test_mech_omnipath


class TestOmnipathThread(test_mech_omnipath.TestOmniPathMechanismDriver):

    def setUp(self):
        super(TestOmnipathThread, self).setUp()
        self.fabric_agent = self.mech_driver.opafmvf.cli

    def test__commit_and_reload(self):
        with mock.patch.object(self.fabric_agent,
                               'osfa_management_commands') as mock_mgt:
            self.mech_driver.omnipath_thread._commit_and_reload()
            self.assertEqual(mock_mgt.call_count, 2)
            expected_calls = [mock.call("commit"),
                              mock.call("reload")]
            mock_mgt.assert_has_calls(expected_calls, any_order=True)

    @mock.patch.object(omnipath_thread.OmniPathThread,
                       '_commit_and_reload')
    def test_sync_omnipath_operations(self, mock_cr):
        with mock.patch.object(fabric_agent.FabricAgentCLI,
                               'execute_command',
                               return_value=0) as mock_exec:
            context_n = self._get_fake_network_context()
            self.mech_driver.create_network_precommit(context_n)
            res = opadbapi.get_all_entries_by_state(context_n,
                                                    'pending')
            self.mech_driver.omnipath_thread.sync_omnipath_operations()
            if res:
                mock_exec.assert_called()
            context_p = self._get_fake_port_context()
            self.mech_driver.create_port_precommit(context_p)
            resp = opadbapi.get_all_entries_by_state(context_p,
                                                     'waiting')
            self.mech_driver.omnipath_thread.sync_omnipath_operations()
            if resp:
                self.assertTrue(mock_exec.called)

    def test__process_entry(self):
        context_n = self._get_fake_network_context()
        context_p = self._get_fake_port_context()
        with mock.patch.object(fabric_agent.FabricAgentCLI,
                               'execute_command') as mock_exec:
            self.mech_driver.create_network_precommit(context_n)
            self.mech_driver.delete_network_precommit(context_n)
            res = opadbapi.get_all_entries_by_state(context_n,
                                                    'pending')
            self.mech_driver.create_port_precommit(context_p)
            resp = opadbapi.get_all_entries_by_state(context_p,
                                                     'waiting')
            self.mech_driver.omnipath_thread._process_entry(
                context_n, res[0])
            self.mech_driver.omnipath_thread._process_entry(
                context_p, resp[0])
            self.assertEqual(mock_exec.call_count, 4)

    def test__process_entry_guid_format(self):
        context_p = self._get_fake_port_context()
        with mock.patch.object(fabric_agent.FabricAgentCLI,
                               'osfa_config_commands') as mock_exec:
            self.mech_driver.create_port_precommit(context_p)
            resp = opadbapi.get_all_entries_by_state(context_p,
                                                     'waiting')
            self.mech_driver.omnipath_thread._process_entry(
                context_p, resp[0])
            guid = resp[0]['data']['port_guid']
            netw = resp[0]['data']['network_id']
            mock_exec.assert_called_with(
                'add', netw, "0x{}".format(guid))  # 0x must be added

    def _create_foo_ports(self):
        foo_port1 = mock.MagicMock()
        foo_port2 = mock.MagicMock()
        foo_port1.data = {'id': '234',
                          'operation': 'add',
                          'network_id': 'net-234',
                          'guid': 'fake_guid'}
        foo_port1.resource_uuid = 'fake_uuid'
        foo_port2.data = {'id': '456',
                          'operation': 'delete',
                          'network_id': 'net-456',
                          'guid': 'bogus_guid'}
        foo_port1.resource_uuid = 'fake_uuid'
        foo_port2.resource_uuid = 'bogus_uuid'
        return [foo_port1, foo_port2]

    def test__prepare_ports_batch(self):
        foo_waiting_ports = self._create_foo_ports()
        self.assertEqual(
            self.mech_driver.omnipath_thread._prepare_ports_batch(
                foo_waiting_ports), ({'net-234': [('fake_guid',
                                                   'fake_uuid')]},
                                     {'net-456': [('bogus_guid',
                                                   'bogus_uuid')]}))

    def test_process_port_add(self):
        context = self._get_fake_port_context()
        foo_waiting_ports = mock.MagicMock()
        foo_waiting_ports.data = {'id': '123',
                                  'operation': 'add',
                                  'network_id': 'net-123',
                                  'guid': 'y_guid'}
        foo_waiting_ports.resource_uuid = 'y_uuid'
        foo_ports = [foo_waiting_ports]
        with mock.patch.object(fabric_agent.FabricAgentCLI,
                               "execute_command",
                               return_value=0):
            self.assertEqual(
                self.mech_driver.omnipath_thread._process_port_batch(
                    context, foo_ports), 1)

    def test_process_port_batch_del(self):
        context = self._get_fake_port_context()
        foo_waiting_ports = mock.MagicMock()
        foo_waiting_ports.data = {'id': '456',
                                  'operation': 'delete',
                                  'network_id': 'net-456',
                                  'guid': 'x_guid'}
        foo_waiting_ports.resource_uuid = 'x_uuid'
        foo_ports = [foo_waiting_ports]
        with mock.patch.object(fabric_agent.FabricAgentCLI,
                               "execute_command",
                               return_value=0):
            self.assertEqual(
                self.mech_driver.omnipath_thread._process_port_batch(
                    context, foo_ports), 1)

    def test__process_port_batch_guid_format(self):
        context = self._get_fake_port_context()
        waiting_ports = self._create_foo_ports()
        with mock.patch.object(fabric_agent.FabricAgentCLI,
                               'osfa_config_commands') as mock_exec:
            self.mech_driver.omnipath_thread._process_port_batch(
                context, waiting_ports)
            mock_exec.assert_any_call('add', 'net-234', '0xfake_guid')
            mock_exec.assert_any_call('remove', 'net-456', '0xbogus_guid')
            self.assertEqual(mock_exec.call_count, 2)



# Copyright 2019 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add_omnipath_journal

Revision ID: fe3470e95eac
Revises: initial_branchpoint
Create Date: 2019-02-05 23:05:47.753557

"""


from alembic import op
from neutron_lib.db import sqlalchemytypes
from oslo_utils import timeutils
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'fe3470e95eac'
down_revision = 'initial_branchpoint'


def upgrade():
    op.create_table(
        'omnipath_revision_journal',
        sa.Column('standard_attr_id', sa.BigInteger, nullable=True,
                  unique=True),
        sa.Column('resource_uuid', sa.String(36), primary_key=True),
        sa.Column('resource_type', sa.String(36), primary_key=True),
        sa.Column('revision_number', sa.BigInteger, nullable=False,
                  default=-1),
        sa.Column('created_at', sqlalchemytypes.TruncatedDateTime,
                  default=timeutils.utcnow, nullable=False),
        sa.Column('updated_at', sqlalchemytypes.TruncatedDateTime,
                  onupdate=timeutils.utcnow),
        sa.Column('data', sa.PickleType, nullable=True),
        sa.Column('state', sa.Enum('pending', 'failed', 'completed',
                                   'waiting', name='state'),
                  nullable=False, default='pending'),
        sa.ForeignKeyConstraint(
            ['standard_attr_id'], ['standardattributes.id'],
            ondelete='SET NULL')
    )



# Copyright (c) 2019 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import paramiko

from neutron_lib import exceptions

from neutron._i18n import _


class FabricAgentCLIError(paramiko.SSHException):
    """Fabric Agent exception"""
    message = _("Unable to connect Fabric Agent.")


class FabricAgentUnknownCommandError(exceptions.NeutronException):
    """Fabric agent unknown command exception"""
    message = _("fabric agent doesn't recognize the instructed command")


class OmnipathResourceExists(exceptions.InUse):
    """Unable to delete resource """
    message = _("Unable to delete resource %(resource)s. "
                "with id %(res_id)s.")


class FabricAgentBadOperation(exceptions.BadRequest):
    """Agent operation exception"""
    message = ("Bad request resource request failed. ")


class NetworkCreateException(exceptions.NeutronException):
    """Fabric Agent network create exception"""
    message = ("Unable to create network %(net_id). ")



# Copyright (c) 2019 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import paramiko

from oslo_config import cfg as config
from oslo_log import log as logging

from omnipath.common import constants as const
from omnipath.common import omnipath_conf
from omnipath.common import omnipath_exceptions

LOG = logging.getLogger(__name__)


class FabricAgentCLI(object):
    def __init__(self):
        self._agent_hostname = None
        self._agent_username = None
        self._agent_key_path = None
        config.CONF.register_opts(omnipath_conf.omnipath_opts,
                                  "ml2_omnipath")

        self._read_config()

        self.client = paramiko.SSHClient()

    def _read_config(self):
        self._agent_hostname = config.CONF.ml2_omnipath.ip_address
        LOG.info("Fabric Agent IP address: %s", self._agent_hostname)
        self._agent_username = config.CONF.ml2_omnipath.username
        self._agent_key_path = config.CONF.ml2_omnipath.ssh_key

    def connect(self):
        try:
            key = paramiko.RSAKey.from_private_key_file(self._agent_key_path)
            self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            self.client.connect(
                self._agent_hostname, port=22,
                username=self._agent_username, pkey=key)
        except omnipath_exceptions.FabricAgentCLIError:
            LOG.error("Error connecting to Omnipath FM")

    def execute_command(self, command):
        self.connect()
        exec_engine = self.client.get_transport().open_session()
        exec_engine.exec_command(command)
        LOG.debug("OMNIPATH Command dispatched %s", command)
        exec_status = exec_engine.recv_exit_status()
        LOG.debug("OMNIPATH Command exit status %s", exec_status)
        self.client.close()
        return exec_status

    def _prepare_command(self, cmd):
        command = ""
        for bit in cmd:
            command = command + " " + bit
        return command

    def osfa_config_commands(self, command, vf_name, *args):
        try:
            if command == const.OPA_CREATE:
                pkey = "--pkey " + str(args[0])
                cmd = [const.OPA_BINARY, const.OPA_CREATE, vf_name, pkey]
            elif command == const.OPA_DELETE:
                cmd = [const.OPA_BINARY, const.OPA_DELETE, vf_name]
            elif command == const.OPA_ADD:
                cmd = [const.OPA_BINARY, const.OPA_ADD, vf_name,
                       "".join(str(x + " ") for x in args).rstrip()]
            elif command == const.OPA_REMOVE:
                cmd = [const.OPA_BINARY, const.OPA_REMOVE, vf_name,
                       "".join(str(x + " ") for x in args).rstrip()]
            else:
                raise omnipath_exceptions.FabricAgentUnknownCommandError
            final_cmd = self._prepare_command(cmd)
            LOG.debug("OMNIPATH command %s" % final_cmd)
            return self.execute_command(final_cmd)
        except omnipath_exceptions.FabricAgentUnknownCommandError:
            LOG.error(command + " not supported in opafmvf CLI")

    def osfa_query_commands(self, command, vf_name, *args):
        try:
            if command == const.OPA_EXIST:
                cmd = [const.OPA_BINARY, const.OPA_EXIST, vf_name]
            elif command == const.OPA_ISMEMBER:
                cmd = [const.OPA_BINARY, const.OPA_ISMEMBER, vf_name,
                       "".join(str(x + " ") for x in args).rstrip()]
            elif command == const.OPA_ISNOTMEMBER:
                cmd = [const.OPA_BINARY, const.OPA_ISNOTMEMBER, vf_name,
                       "".join(str(x + " ") for x in args).rstrip()]
            else:
                raise omnipath_exceptions.FabricAgentUnknownCommandError
            final_cmd = self._prepare_command(cmd)
            LOG.debug("OMNIPATH command %s" % final_cmd)
            return self.execute_command(final_cmd)
        except omnipath_exceptions.FabricAgentUnknownCommandError:
            LOG.error(command + " not supported in opafmvf CLI")

    def osfa_management_commands(self, command):
        try:
            if command == const.OPA_RESET:
                cmd = [const.OPA_BINARY, const.OPA_RESET]
            elif command == const.OPA_COMMIT:
                cmd = [const.OPA_BINARY, const.OPA_COMMIT, "-f"]
            elif command == const.OPA_RELOAD:
                cmd = [const.OPA_BINARY, const.OPA_RELOAD]
            elif command == const.OPA_RESTART:
                cmd = [const.OPA_BINARY, const.OPA_RESTART]
            elif command == const.OPA_ABORT:
                cmd = ["killall -9", const.OPA_BINARY]
            else:
                raise omnipath_exceptions.FabricAgentUnknownCommandError
            final_cmd = self._prepare_command(cmd)
            LOG.debug("OMNIPATH command %s" % final_cmd)
            return self.execute_command(final_cmd)
        except omnipath_exceptions.FabricAgentUnknownCommandError:
            LOG.error(command + " not supported in opafmvf CLI")


class FabricAgentClient(object):
    def __init__(self):
        self.cli = FabricAgentCLI()

    def get_port_status(self, vf_name, guid):
        """Get status of port

        :param vf_name: Name of the VF
        :param guid: ID of the physical server
        :return: bind status
        """

        query_status = self.cli.osfa_query_commands(
            const.OPA_ISMEMBER, vf_name, guid)
        if query_status == 0:
            return "UP"
        else:
            return "DOWN"



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import random

import mock

from neutron.objects import ports

from oslo_utils import uuidutils

from omnipath.db import api as db_api
from omnipath.tests.unit import base


class TestOmniPathMechanismDriver(base.TestOmniPathBase, base.DBTestCase):

    def setUp(self):
        super(TestOmniPathMechanismDriver, self).setUp()

    def _get_fake_network_context(self):
        current = {'status': 'ACTIVE',
                   'subnets': [],
                   'name': 'net1',
                   'provider:physical_network': None,
                   'admin_state_up': True,
                   'tenant_id': 'test-tenant',
                   'provider:network_type': 'vlan',
                   'shared': False,
                   'id': uuidutils.generate_uuid(),
                   'provider:segmentation_id': random.randint(2, 2000)}
        context = mock.Mock(current=current)
        context.session = self.session
        context._plugin_context = self.db_context
        return context

    def _get_fake_port_context(self):
        fake_net = self._get_fake_network_context()
        current = {'status': 'DOWN',
                   'binding:host_id': '',
                   'allowed_address_pairs': [],
                   'device_owner': 'fake_owner',
                   'binding:profile': {'guid': 'fake_guid'},
                   'fixed_ips': [{
                       'subnet_id': uuidutils.generate_uuid()}],
                   'id': uuidutils.generate_uuid(),
                   'device_id': 'fake_device',
                   'port_guid': 'fake_guid',
                   'name': '',
                   'admin_state_up': True,
                   'network_id': fake_net.current['id'],
                   'tenant_id': fake_net.current['tenant_id'],
                   'binding:vif_details': {},
                   'binding:vnic_type': 'baremetal',
                   'binding:vif_type': 'unbound',
                   'mac_address': 'fake_mac',
                   'extra_dhcp_opts': [{
                       'opt_name': '61',
                       'opt_value': '0xfe80000000000000001122334455667f'}]}
        context = mock.Mock(current=current)
        context.network = fake_net
        context.session = self.session
        context._plugin_context = self.db_context
        return context

    def _assert_resource_row(self, context, res_op, res_type):
        result = db_api.get_resource_row(context, context.current['id'],
                                         res_type)
        self.assertEqual(result.data, context.current)
        self.assertEqual(result.data['operation'], res_op)

    def test_network_create_precommit(self):
        context = self._get_fake_network_context()
        self.mech_driver.create_network_precommit(context)
        self._assert_resource_row(context, "create", "network")

    def test_network_delete_precommit(self):
        context = self._get_fake_network_context()
        self.mech_driver.create_network_precommit(context)
        self.mech_driver.delete_network_precommit(context)
        self._assert_resource_row(context, "delete", "network")

    def test_port_create_precommit(self):
        context = self._get_fake_port_context()
        self.mech_driver.create_port_precommit(context)
        self._assert_resource_row(context, "create", "port")

    def test_port_delete_precommit(self):
        context = self._get_fake_port_context()
        self.mech_driver.create_port_precommit(context)
        self.mech_driver.delete_port_precommit(context)
        self._assert_resource_row(context, "delete", "port")

    def test__is_port_supported(self):
        context = self._get_fake_port_context()
        is_supported = self.mech_driver._is_port_supported(context.current)
        self.assertTrue(is_supported)

    def test__is_port_supported_false(self):
        context = self._get_fake_port_context()
        context.current['binding:vnic_type'] = 'not_baremetal'
        is_supported = self.mech_driver._is_port_supported(context.current)
        self.assertFalse(is_supported)

    def test_db_multiple_rows_get(self):
        ctx_net1 = self._get_fake_network_context()
        ctx_net2 = self._get_fake_network_context()
        self.mech_driver.create_network_precommit(ctx_net1)
        self.mech_driver.create_network_precommit(ctx_net2)
        res = db_api.get_all_entries_by_state(ctx_net2, "pending")
        self.assertEqual(2, len(res))
        db_api.update_multiple_rows(ctx_net2, "completed",
                                    [ctx_net1.current['id'],
                                     ctx_net2.current['id']])
        res2 = db_api.get_all_entries_by_state(ctx_net2, "completed")
        self.assertEqual(2, len(res2))
        for row in res2:
            self.assertEqual("completed", row.state)

    def test_all_postcommits(self):
        with mock.patch.object(
            self.mech_driver.omnipath_thread, "set_sync_event") \
                as mock_sync:
            self.mech_driver.create_network_postcommit(mock.ANY)
            self.mech_driver.delete_network_postcommit(mock.ANY)
            self.mech_driver.create_port_postcommit(mock.ANY)
            self.mech_driver.delete_port_postcommit(mock.ANY)
            # call count should be 3 since create_port doesn't set_sync_event
            self.assertEqual(3, mock_sync.call_count)

    @mock.patch('neutron.db.provisioning_blocks.provisioning_complete')
    def test_bind_port(self, mock_pb):
        ctx_port = self._get_fake_port_context()
        ctx_port.segments_to_bind = [{'id': 'fake_segment'}]
        ctx_port.set_binding = mock.Mock()
        with mock.patch.object(
            self.mech_driver.omnipath_thread, "set_sync_event") \
                as mock_bind:
            self.mech_driver.create_port_precommit(ctx_port)
            self.mech_driver.bind_port(ctx_port)
        self.assertTrue(mock_bind.called)
        self.assertTrue(mock_pb.called)

    @mock.patch.object(ports.Port, 'update_object')
    def test_update_port_status_db(self, port_update):
        ctx_port = self._get_fake_port_context()
        self.mech_driver.update_port_status_db(ctx_port,
                                               'fake_port_id',
                                               'pending')
        self.assertTrue(port_update.called)

    def test__unplug_port_from_network(self):
        ctx_port = self._get_fake_port_context()
        with mock.patch.object(
                self.mech_driver.omnipath_thread, "set_sync_event") as msync:
            self.mech_driver._unplug_port_from_network(ctx_port)
            self.assertEqual(1, msync.call_count)
